import os.path

import pytest

import salt.modules.archive as archive
from salt.exceptions import CommandNotFoundError
from tests.support.mock import MagicMock, patch


class ZipFileMock(MagicMock):
    def __init__(self, files=None, **kwargs):  # pylint: disable=W0231
        if files is None:
            files = ["salt"]
        MagicMock.__init__(self, **kwargs)
        self._files = files
        self.external_attr = 0o0777 << 16

    def namelist(self):
        return self._files


@pytest.fixture
def configure_loader_modules():
    return {archive: {"__grains__": {"id": 0}}}


def test_tar():
    with patch("glob.glob", lambda pathname: [pathname]):
        with patch("salt.utils.path.which", lambda exe: exe):
            mock = MagicMock(return_value="salt")
            with patch.dict(archive.__salt__, {"cmd.run": mock}):
                ret = archive.tar(
                    "-zcvf",
                    "foo.tar",
                    ["/tmp/something-to-compress-1", "/tmp/something-to-compress-2"],
                    template=None,
                )
                assert ["salt"] == ret
                mock.assert_called_once_with(
                    [
                        "tar",
                        "-zcvf",
                        "foo.tar",
                        "/tmp/something-to-compress-1",
                        "/tmp/something-to-compress-2",
                    ],
                    runas=None,
                    python_shell=False,
                    template=None,
                    cwd=None,
                )

            mock = MagicMock(return_value="salt")
            with patch.dict(archive.__salt__, {"cmd.run": mock}):
                ret = archive.tar(
                    "-zcvf",
                    "foo.tar",
                    "/tmp/something-to-compress-1,/tmp/something-to-compress-2",
                    template=None,
                )
                assert ["salt"] == ret
                mock.assert_called_once_with(
                    [
                        "tar",
                        "-zcvf",
                        "foo.tar",
                        "/tmp/something-to-compress-1",
                        "/tmp/something-to-compress-2",
                    ],
                    runas=None,
                    python_shell=False,
                    template=None,
                    cwd=None,
                )


def test_tar_raises_exception_if_not_found():
    with patch("salt.utils.path.which", lambda exe: None):
        mock = MagicMock(return_value="salt")
        with patch.dict(archive.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandNotFoundError):
                archive.tar("zxvf", "foo.tar", "/tmp/something-to-compress")
            assert not mock.called


def test_gzip():
    mock = MagicMock(return_value="salt")
    with patch.dict(archive.__salt__, {"cmd.run": mock}):
        with patch("salt.utils.path.which", lambda exe: exe):
            ret = archive.gzip("/tmp/something-to-compress")
            assert ["salt"] == ret
            mock.assert_called_once_with(
                ["gzip", "/tmp/something-to-compress"],
                runas=None,
                python_shell=False,
                template=None,
            )


def test_gzip_raises_exception_if_not_found():
    mock = MagicMock(return_value="salt")
    with patch.dict(archive.__salt__, {"cmd.run": mock}):
        with patch("salt.utils.path.which", lambda exe: None):
            with pytest.raises(CommandNotFoundError):
                archive.gzip("/tmp/something-to-compress")
            assert not mock.called


def test_gunzip():
    mock = MagicMock(return_value="salt")
    with patch.dict(archive.__salt__, {"cmd.run": mock}):
        with patch("salt.utils.path.which", lambda exe: exe):
            ret = archive.gunzip("/tmp/something-to-decompress.tar.gz")
            assert ["salt"] == ret
            mock.assert_called_once_with(
                ["gunzip", "/tmp/something-to-decompress.tar.gz"],
                runas=None,
                python_shell=False,
                template=None,
            )


def test_gunzip_raises_exception_if_not_found():
    mock = MagicMock(return_value="salt")
    with patch.dict(archive.__salt__, {"cmd.run": mock}):
        with patch("salt.utils.path.which", lambda exe: None):
            with pytest.raises(CommandNotFoundError):
                archive.gunzip("/tmp/something-to-decompress.tar.gz")
            assert not mock.called


def test_cmd_zip():
    with patch("glob.glob", lambda pathname: [pathname]):
        with patch("salt.utils.path.which", lambda exe: exe):
            mock = MagicMock(return_value="salt")
            with patch.dict(archive.__salt__, {"cmd.run": mock}):
                ret = archive.cmd_zip(
                    "/tmp/salt.{{grains.id}}.zip",
                    "/tmp/tmpePe8yO,/tmp/tmpLeSw1A",
                    template="jinja",
                )
                assert ["salt"] == ret
                mock.assert_called_once_with(
                    [
                        "zip",
                        "-r",
                        "/tmp/salt.{{grains.id}}.zip",
                        "/tmp/tmpePe8yO",
                        "/tmp/tmpLeSw1A",
                    ],
                    runas=None,
                    python_shell=False,
                    template="jinja",
                    cwd=None,
                )

            mock = MagicMock(return_value="salt")
            with patch.dict(archive.__salt__, {"cmd.run": mock}):
                ret = archive.cmd_zip(
                    "/tmp/salt.{{grains.id}}.zip",
                    ["/tmp/tmpePe8yO", "/tmp/tmpLeSw1A"],
                    template="jinja",
                )
                assert ["salt"] == ret
                mock.assert_called_once_with(
                    [
                        "zip",
                        "-r",
                        "/tmp/salt.{{grains.id}}.zip",
                        "/tmp/tmpePe8yO",
                        "/tmp/tmpLeSw1A",
                    ],
                    runas=None,
                    python_shell=False,
                    template="jinja",
                    cwd=None,
                )


def test_zip():
    with patch("glob.glob", lambda pathname: [pathname]):
        with patch.multiple(
            os.path,
            **{
                "isdir": MagicMock(return_value=False),
                "exists": MagicMock(return_value=True),
            }
        ):
            with patch("zipfile.ZipFile", MagicMock()):
                ret = archive.zip_(
                    "/tmp/salt.{{grains.id}}.zip",
                    "/tmp/tmpePe8yO,/tmp/tmpLeSw1A",
                    template="jinja",
                )
                expected = [
                    os.path.join("tmp", "tmpePe8yO"),
                    os.path.join("tmp", "tmpLeSw1A"),
                ]
                assert expected == ret


def test_zip_raises_exception_if_not_found():
    mock = MagicMock(return_value="salt")
    with patch.dict(archive.__salt__, {"cmd.run": mock}):
        with patch("salt.utils.path.which", lambda exe: None):
            with pytest.raises(CommandNotFoundError):
                archive.cmd_zip(
                    "/tmp/salt.{{grains.id}}.zip",
                    "/tmp/tmpePe8yO,/tmp/tmpLeSw1A",
                    template="jinja",
                )
            assert not mock.called


def test_cmd_unzip():
    def _get_mock():
        """
        Create a new MagicMock for each scenario in this test, so that
        assert_called_once_with doesn't complain that the same mock object
        is called more than once.
        """
        return MagicMock(
            return_value={"stdout": "salt", "stderr": "", "pid": 12345, "retcode": 0}
        )

    with patch("salt.utils.path.which", lambda exe: exe):

        mock = _get_mock()
        with patch.dict(archive.__salt__, {"cmd.run_all": mock}):
            ret = archive.cmd_unzip(
                "/tmp/salt.{{grains.id}}.zip",
                "/tmp/dest",
                excludes="/tmp/tmpePe8yO,/tmp/tmpLeSw1A",
                template="jinja",
            )
            assert ["salt"] == ret
            mock.assert_called_once_with(
                [
                    "unzip",
                    "/tmp/salt.{{grains.id}}.zip",
                    "-d",
                    "/tmp/dest",
                    "-x",
                    "/tmp/tmpePe8yO",
                    "/tmp/tmpLeSw1A",
                ],
                output_loglevel="debug",
                python_shell=False,
                redirect_stderr=True,
                runas=None,
                template="jinja",
            )

        mock = _get_mock()
        with patch.dict(archive.__salt__, {"cmd.run_all": mock}):
            ret = archive.cmd_unzip(
                "/tmp/salt.{{grains.id}}.zip",
                "/tmp/dest",
                excludes=["/tmp/tmpePe8yO", "/tmp/tmpLeSw1A"],
                template="jinja",
            )
            assert ["salt"] == ret
            mock.assert_called_once_with(
                [
                    "unzip",
                    "/tmp/salt.{{grains.id}}.zip",
                    "-d",
                    "/tmp/dest",
                    "-x",
                    "/tmp/tmpePe8yO",
                    "/tmp/tmpLeSw1A",
                ],
                output_loglevel="debug",
                python_shell=False,
                redirect_stderr=True,
                runas=None,
                template="jinja",
            )

        mock = _get_mock()
        with patch.dict(archive.__salt__, {"cmd.run_all": mock}):
            ret = archive.cmd_unzip(
                "/tmp/salt.{{grains.id}}.zip",
                "/tmp/dest",
                excludes="/tmp/tmpePe8yO,/tmp/tmpLeSw1A",
                template="jinja",
                options="-fo",
            )
            assert ["salt"] == ret
            mock.assert_called_once_with(
                [
                    "unzip",
                    "-fo",
                    "/tmp/salt.{{grains.id}}.zip",
                    "-d",
                    "/tmp/dest",
                    "-x",
                    "/tmp/tmpePe8yO",
                    "/tmp/tmpLeSw1A",
                ],
                output_loglevel="debug",
                python_shell=False,
                redirect_stderr=True,
                runas=None,
                template="jinja",
            )

        mock = _get_mock()
        with patch.dict(archive.__salt__, {"cmd.run_all": mock}):
            ret = archive.cmd_unzip(
                "/tmp/salt.{{grains.id}}.zip",
                "/tmp/dest",
                excludes=["/tmp/tmpePe8yO", "/tmp/tmpLeSw1A"],
                template="jinja",
                options="-fo",
            )
            assert ["salt"] == ret
            mock.assert_called_once_with(
                [
                    "unzip",
                    "-fo",
                    "/tmp/salt.{{grains.id}}.zip",
                    "-d",
                    "/tmp/dest",
                    "-x",
                    "/tmp/tmpePe8yO",
                    "/tmp/tmpLeSw1A",
                ],
                output_loglevel="debug",
                python_shell=False,
                redirect_stderr=True,
                runas=None,
                template="jinja",
            )

        mock = _get_mock()
        with patch.dict(archive.__salt__, {"cmd.run_all": mock}):
            ret = archive.cmd_unzip(
                "/tmp/salt.{{grains.id}}.zip",
                "/tmp/dest",
                excludes=["/tmp/tmpePe8yO", "/tmp/tmpLeSw1A"],
                template="jinja",
                options="-fo",
                password="asdf",
            )
            assert ["salt"] == ret
            mock.assert_called_once_with(
                [
                    "unzip",
                    "-P",
                    "asdf",
                    "-fo",
                    "/tmp/salt.{{grains.id}}.zip",
                    "-d",
                    "/tmp/dest",
                    "-x",
                    "/tmp/tmpePe8yO",
                    "/tmp/tmpLeSw1A",
                ],
                output_loglevel="quiet",
                python_shell=False,
                redirect_stderr=True,
                runas=None,
                template="jinja",
            )


def test_unzip():
    mock = ZipFileMock()
    with patch("zipfile.ZipFile", mock):
        ret = archive.unzip(
            "/tmp/salt.{{grains.id}}.zip",
            "/tmp/dest",
            excludes="/tmp/tmpePe8yO,/tmp/tmpLeSw1A",
            template="jinja",
            extract_perms=False,
        )
        assert ["salt"] == ret


def test_unzip_password():
    mock = ZipFileMock()
    target = "/tmp/salt.zip"
    dest = "/tmp/dest"
    # This is the file inside target zip file, returned by ZipFileMock.namelist
    expected_file_name = "salt"
    with patch("zipfile.ZipFile", mock):
        ret = archive.unzip(target, dest, password="spongebob", extract_perms=False)
        assert [expected_file_name] == ret
        # mock needs to be mock() because contextlib.closing appears to actually
        # call the mock
        mock().extract.assert_called_once_with(expected_file_name, dest, b"spongebob")


def test_unzip_password_bytestring():
    mock = ZipFileMock()
    target = "/tmp/salt.zip"
    dest = "/tmp/dest"
    # This is the file inside target zip file, returned by ZipFileMock.namelist
    expected_file_name = "salt"
    with patch("zipfile.ZipFile", mock):
        ret = archive.unzip(target, dest, password=b"spongebob", extract_perms=False)
        assert [expected_file_name] == ret
        # mock needs to be mock() because contextlib.closing appears to actually
        # call the mock
        mock().extract.assert_called_once_with(expected_file_name, dest, b"spongebob")


def test_unzip_password_integer():
    mock = ZipFileMock()
    target = "/tmp/salt.zip"
    dest = "/tmp/dest"
    # This is the file inside target zip file, returned by ZipFileMock.namelist
    expected_file_name = "salt"
    with patch("zipfile.ZipFile", mock):
        ret = archive.unzip(target, dest, password=12345, extract_perms=False)
        assert [expected_file_name] == ret
        # mock needs to be mock() because contextlib.closing appears to actually
        # call the mock
        mock().extract.assert_called_once_with(expected_file_name, dest, b"12345")


def test_unzip_raises_exception_if_not_found():
    mock = MagicMock(return_value="salt")
    with patch.dict(archive.__salt__, {"cmd.run": mock}):
        with patch("salt.utils.path.which", lambda exe: None):
            with pytest.raises(CommandNotFoundError):
                archive.cmd_unzip(
                    "/tmp/salt.{{grains.id}}.zip",
                    "/tmp/dest",
                    excludes="/tmp/tmpePe8yO,/tmp/tmpLeSw1A",
                    template="jinja",
                )
            assert not mock.called


def test_rar():
    with patch("glob.glob", lambda pathname: [pathname]):
        with patch("salt.utils.path.which", lambda exe: exe):
            mock = MagicMock(return_value="salt")
            with patch.dict(archive.__salt__, {"cmd.run": mock}):
                ret = archive.rar(
                    "/tmp/rarfile.rar", "/tmp/sourcefile1,/tmp/sourcefile2"
                )
                assert ["salt"] == ret
                mock.assert_called_once_with(
                    [
                        "rar",
                        "a",
                        "-idp",
                        "/tmp/rarfile.rar",
                        "/tmp/sourcefile1",
                        "/tmp/sourcefile2",
                    ],
                    runas=None,
                    python_shell=False,
                    template=None,
                    cwd=None,
                )

            mock = MagicMock(return_value="salt")
            with patch.dict(archive.__salt__, {"cmd.run": mock}):
                ret = archive.rar(
                    "/tmp/rarfile.rar", ["/tmp/sourcefile1", "/tmp/sourcefile2"]
                )
                assert ["salt"] == ret
                mock.assert_called_once_with(
                    [
                        "rar",
                        "a",
                        "-idp",
                        "/tmp/rarfile.rar",
                        "/tmp/sourcefile1",
                        "/tmp/sourcefile2",
                    ],
                    runas=None,
                    python_shell=False,
                    template=None,
                    cwd=None,
                )


def test_rar_raises_exception_if_not_found():
    mock = MagicMock(return_value="salt")
    with patch.dict(archive.__salt__, {"cmd.run": mock}):
        with patch("salt.utils.path.which", lambda exe: None):
            with pytest.raises(CommandNotFoundError):
                archive.rar("/tmp/rarfile.rar", "/tmp/sourcefile1,/tmp/sourcefile2")
            assert not mock.called


@pytest.mark.skip_if_binaries_missing("unrar", "rar", reason="unrar not installed")
def test_unrar():
    with patch(
        "salt.utils.path.which_bin",
        lambda exe: exe[0] if isinstance(exe, (list, tuple)) else exe,
    ):
        with patch(
            "salt.utils.path.which",
            lambda exe: exe[0] if isinstance(exe, (list, tuple)) else exe,
        ):
            mock = MagicMock(return_value="salt")
            with patch.dict(archive.__salt__, {"cmd.run": mock}):
                ret = archive.unrar(
                    "/tmp/rarfile.rar", "/home/strongbad/", excludes="file_1,file_2"
                )
                assert ["salt"] == ret
                mock.assert_called_once_with(
                    [
                        "unrar",
                        "x",
                        "-idp",
                        "/tmp/rarfile.rar",
                        "-x",
                        "file_1",
                        "-x",
                        "file_2",
                        "/home/strongbad/",
                    ],
                    runas=None,
                    python_shell=False,
                    template=None,
                )

            mock = MagicMock(return_value="salt")
            with patch.dict(archive.__salt__, {"cmd.run": mock}):
                ret = archive.unrar(
                    "/tmp/rarfile.rar",
                    "/home/strongbad/",
                    excludes=["file_1", "file_2"],
                )
                assert ["salt"] == ret
                mock.assert_called_once_with(
                    [
                        "unrar",
                        "x",
                        "-idp",
                        "/tmp/rarfile.rar",
                        "-x",
                        "file_1",
                        "-x",
                        "file_2",
                        "/home/strongbad/",
                    ],
                    runas=None,
                    python_shell=False,
                    template=None,
                )


def test_unrar_raises_exception_if_not_found():
    with patch("salt.utils.path.which_bin", lambda exe: None):
        mock = MagicMock(return_value="salt")
        with patch.dict(archive.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandNotFoundError):
                archive.unrar(
                    "/tmp/rarfile.rar",
                    "/home/strongbad/",
                )
            assert not mock.called


def test_trim_files():
    with patch("salt.utils.path.which_bin", lambda exe: exe):
        source = "file.tar.gz"
        tmp_dir = "temp"
        files = [
            "\n".join(["file1"] * 200),
            "\n".join(["file2"] * 200),
            "this\nthat\nother",
            "this\nthat\nother",
            "this\nthat\nother",
        ]
        trim_opt = [True, False, 3, 1, 0]
        trim_100 = ["file1"] * 100
        trim_100.append("List trimmed after 100 files.")
        expected = [
            trim_100,
            ["file2"] * 200,
            ["this", "that", "other"],
            ["this", "List trimmed after 1 files."],
            ["List trimmed after 0 files."],
        ]

        for test_files, test_trim_opts, test_expected in zip(files, trim_opt, expected):
            with patch.dict(
                archive.__salt__, {"cmd.run": MagicMock(return_value=test_files)}
            ):
                ret = archive.unrar(source, tmp_dir, trim_output=test_trim_opts)
                assert ret == test_expected



"""
Integration tests for the idem execution module
"""

from contextlib import contextmanager

import pytest

import salt.utils.idem as idem
import salt.utils.path

pytestmark = [
    pytest.mark.skipif(not idem.HAS_POP[0], reason=idem.HAS_POP[1]),
]


@pytest.mark.skipif(not salt.utils.path.which("idem"), reason="idem is not installed")
@contextmanager
def test_exec(salt_call_cli):
    ret = salt_call_cli.run("--local", "idem.exec", "test.ping")
    assert ret.data is True



import pytest

import salt.states.mac_assistive as assistive
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {assistive: {}}


def test_installed():
    """
    Test installing a bundle ID as being allowed to run with assistive access
    """
    expected = {
        "changes": {},
        "comment": "Installed com.apple.Chess into the assistive access panel",
        "name": "com.apple.Chess",
        "result": True,
    }

    installed_mock = MagicMock(return_value=False)
    install_mock = MagicMock()

    with patch.dict(
        assistive.__salt__,
        {"assistive.installed": installed_mock, "assistive.install": install_mock},
    ):
        out = assistive.installed("com.apple.Chess")
        installed_mock.assert_called_once_with("com.apple.Chess")
        install_mock.assert_called_once_with("com.apple.Chess", True)
        assert out == expected


def test_installed_not_enabled():
    """
    Test installing a bundle ID as being allowed to run with disabled assistive access
    """
    expected = {
        "changes": {},
        "comment": "Updated enable to True",
        "name": "com.apple.Chess",
        "result": True,
    }

    installed_mock = MagicMock(return_value=True)
    install_mock = MagicMock()
    enabled_mock = MagicMock(return_value=False)
    enable_mock = MagicMock()

    with patch.dict(
        assistive.__salt__,
        {
            "assistive.installed": installed_mock,
            "assistive.install": install_mock,
            "assistive.enabled": enabled_mock,
            "assistive.enable": enable_mock,
        },
    ):
        out = assistive.installed("com.apple.Chess")
        enabled_mock.assert_called_once_with("com.apple.Chess")
        enable_mock.assert_called_once_with("com.apple.Chess", True)
        assert not install_mock.called
        assert out == expected


def test_installed_enabled():
    """
    Test enabling an already enabled bundle ID
    """
    expected = {
        "changes": {},
        "comment": "Already in the correct state",
        "name": "com.apple.Chess",
        "result": True,
    }

    installed_mock = MagicMock(return_value=True)
    install_mock = MagicMock()
    enabled_mock = MagicMock(return_value=True)
    enable_mock = MagicMock()

    with patch.dict(
        assistive.__salt__,
        {
            "assistive.installed": installed_mock,
            "assistive.install": install_mock,
            "assistive.enabled": enabled_mock,
            "assistive.enable": enable_mock,
        },
    ):
        out = assistive.installed("com.apple.Chess")
        enabled_mock.assert_called_once_with("com.apple.Chess")
        assert not enable_mock.called
        assert not install_mock.called
        assert out == expected


def test_installed_not_disabled():
    """
    Test disabling an enabled and installed bundle ID
    """
    expected = {
        "changes": {},
        "comment": "Updated enable to False",
        "name": "com.apple.Chess",
        "result": True,
    }

    installed_mock = MagicMock(return_value=True)
    install_mock = MagicMock()
    enabled_mock = MagicMock(return_value=True)
    enable_mock = MagicMock()

    with patch.dict(
        assistive.__salt__,
        {
            "assistive.installed": installed_mock,
            "assistive.install": install_mock,
            "assistive.enabled": enabled_mock,
            "assistive.enable": enable_mock,
        },
    ):
        out = assistive.installed("com.apple.Chess", False)
        enabled_mock.assert_called_once_with("com.apple.Chess")
        enable_mock.assert_called_once_with("com.apple.Chess", False)
        assert not install_mock.called
        assert out == expected



#!/usr/bin/env python

import argparse
import modulefinder
import os
import pprint
import sys

import salt.utils.json
import salt.utils.yaml


def parse():
    """
    Parse the arguments
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-r",
        "--root",
        dest="root",
        default=".",
        help="The base code directory to look in",
    )
    parser.add_argument("-i", "--bif", dest="bif", default="site-packages")
    parser.add_argument(
        "-f", "--format", dest="format", choices=("pprint", "yaml"), default="pprint"
    )
    out = parser.parse_args()
    return out.__dict__


def mod_data(opts, full):
    """
    Grab the module's data
    """
    ret = {}
    finder = modulefinder.ModuleFinder()
    try:
        finder.load_file(full)
    except ImportError as exc:
        print(f"ImportError - {full} (Reason: {exc})", file=sys.stderr)
        return ret
    for name, mod in finder.modules.items():
        basemod = name.split(".")[0]
        if basemod in ret:
            continue
        if basemod.startswith("_"):
            continue
        if not mod.__file__:
            continue
        if opts["bif"] not in mod.__file__:
            # Bif - skip
            continue
        if name == os.path.basename(mod.__file__)[:-3]:
            continue
        ret[basemod] = mod.__file__
    for name, err in finder.badmodules.items():
        basemod = name.split(".")[0]
        if basemod in ret:
            continue
        if basemod.startswith("_"):
            continue
        ret[basemod] = err
    return ret


def scan(opts):
    """
    Scan the provided root for python source files
    """
    ret = {}
    for root, dirs, files in os.walk(opts["root"]):
        for fn_ in files:
            full = os.path.join(root, fn_)
            if full.endswith(".py"):
                ret.update(mod_data(opts, full))
    return ret


if __name__ == "__main__":
    opts = parse()
    try:
        scand = scan(opts)
        if opts["format"] == "yaml":
            print(salt.utils.yaml.safe_dump(scand))
        if opts["format"] == "json":
            print(salt.utils.json.dumps(scand))
        else:
            pprint.pprint(scand)
        exit(0)
    except KeyboardInterrupt:
        print("\nInterrupted on user request", file=sys.stderr)
        exit(1)



import subprocess

import pytest

import salt.modules.openscap as openscap
from tests.support.mock import MagicMock, Mock, patch


@pytest.fixture
def policy_file():
    yield "/usr/share/openscap/policy-file-xccdf.xml"


@pytest.fixture
def configure_loader_modules(tmp_path):
    random_temp_dir = tmp_path / "unique"
    random_temp_dir.mkdir()
    with patch("salt.modules.openscap.shutil.rmtree", Mock()), patch(
        "salt.modules.openscap.tempfile.mkdtemp",
        Mock(return_value=str(random_temp_dir)),
    ), patch("salt.modules.openscap.os.path.exists", Mock(return_value=True)):
        yield {openscap: {"__salt__": {"cp.push_dir": MagicMock()}}}


def test_openscap_xccdf_eval_success(policy_file):
    mock_popen = MagicMock(
        return_value=Mock(**{"returncode": 0, "communicate.return_value": ("", "")})
    )
    patch_popen = patch("salt.modules.openscap.subprocess.Popen", mock_popen)
    with patch_popen:
        response = openscap.xccdf(f"eval --profile Default {policy_file}")

        expected_cmd = [
            "oscap",
            "xccdf",
            "eval",
            "--oval-results",
            "--results",
            "results.xml",
            "--report",
            "report.html",
            "--profile",
            "Default",
            policy_file,
        ]
        openscap.subprocess.Popen.assert_called_once_with(
            expected_cmd,
            cwd=openscap.tempfile.mkdtemp.return_value,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        openscap.__salt__["cp.push_dir"].assert_called_once_with(
            openscap.tempfile.mkdtemp.return_value
        )
        assert openscap.shutil.rmtree.call_count == 1
        expected = {
            "upload_dir": openscap.tempfile.mkdtemp.return_value,
            "error": "",
            "success": True,
            "returncode": 0,
        }
        assert response == expected


def test_openscap_xccdf_eval_success_with_failing_rules(policy_file):
    mock_popen = MagicMock(
        return_value=Mock(
            **{"returncode": 2, "communicate.return_value": ("", "some error")}
        )
    )
    patch_popen = patch("salt.modules.openscap.subprocess.Popen", mock_popen)
    with patch_popen:
        response = openscap.xccdf(f"eval --profile Default {policy_file}")

        expected_cmd = [
            "oscap",
            "xccdf",
            "eval",
            "--oval-results",
            "--results",
            "results.xml",
            "--report",
            "report.html",
            "--profile",
            "Default",
            policy_file,
        ]
        openscap.subprocess.Popen.assert_called_once_with(
            expected_cmd,
            cwd=openscap.tempfile.mkdtemp.return_value,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        openscap.__salt__["cp.push_dir"].assert_called_once_with(
            openscap.tempfile.mkdtemp.return_value
        )
        expected = {
            "upload_dir": openscap.tempfile.mkdtemp.return_value,
            "error": "some error",
            "success": True,
            "returncode": 2,
        }
        assert response == expected


def test_openscap_xccdf_eval_fail_no_profile():
    response = openscap.xccdf("eval --param Default /unknown/param")
    error = "the following arguments are required: --profile"
    expected = {
        "error": error,
        "upload_dir": None,
        "success": False,
        "returncode": None,
    }
    assert response == expected


def test_openscap_xccdf_eval_success_ignore_unknown_params():
    mock_popen = MagicMock(
        return_value=Mock(
            **{"returncode": 2, "communicate.return_value": ("", "some error")}
        )
    )
    patch_popen = patch("salt.modules.openscap.subprocess.Popen", mock_popen)
    with patch_popen:
        response = openscap.xccdf("eval --profile Default --param Default /policy/file")
        expected = {
            "upload_dir": openscap.tempfile.mkdtemp.return_value,
            "error": "some error",
            "success": True,
            "returncode": 2,
        }
        assert response == expected
        expected_cmd = [
            "oscap",
            "xccdf",
            "eval",
            "--oval-results",
            "--results",
            "results.xml",
            "--report",
            "report.html",
            "--profile",
            "Default",
            "/policy/file",
        ]
        openscap.subprocess.Popen.assert_called_once_with(
            expected_cmd,
            cwd=openscap.tempfile.mkdtemp.return_value,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )


def test_openscap_xccdf_eval_evaluation_error(policy_file):
    mock_popen = MagicMock(
        return_value=Mock(
            **{
                "returncode": 1,
                "communicate.return_value": ("", "evaluation error"),
            }
        )
    )
    patch_popen = patch("salt.modules.openscap.subprocess.Popen", mock_popen)
    with patch_popen:
        response = openscap.xccdf(f"eval --profile Default {policy_file}")
        expected = {
            "upload_dir": None,
            "error": "evaluation error",
            "success": False,
            "returncode": 1,
        }
        assert response == expected


def test_openscap_xccdf_eval_fail_not_implemented_action(policy_file):
    response = openscap.xccdf(f"info {policy_file}")
    mock_err = "argument action: invalid choice: 'info' (choose from 'eval')"
    expected = {
        "upload_dir": None,
        "error": mock_err,
        "success": False,
        "returncode": None,
    }
    assert response == expected


def test_openscap_xccdf_eval_evaluation_unknown_error(policy_file):
    mock_popen = MagicMock(
        return_value=Mock(
            **{
                "returncode": 255,
                "communicate.return_value": ("", "unknown error"),
            }
        )
    )
    patch_popen = patch("salt.modules.openscap.subprocess.Popen", mock_popen)
    with patch_popen:
        response = openscap.xccdf(f"eval --profile Default {policy_file}")
        expected = {
            "upload_dir": None,
            "error": "unknown error",
            "success": False,
            "returncode": 255,
        }
        assert response == expected


def test_new_openscap_xccdf_eval_success(policy_file):
    with patch(
        "salt.modules.openscap.subprocess.Popen",
        MagicMock(
            return_value=Mock(**{"returncode": 0, "communicate.return_value": ("", "")})
        ),
    ):
        response = openscap.xccdf_eval(
            policy_file,
            profile="Default",
            oval_results=True,
            results="results.xml",
            report="report.html",
        )

        assert openscap.tempfile.mkdtemp.call_count == 1
        expected_cmd = [
            "oscap",
            "xccdf",
            "eval",
            "--oval-results",
            "--results",
            "results.xml",
            "--report",
            "report.html",
            "--profile",
            "Default",
            policy_file,
        ]
        openscap.subprocess.Popen.assert_called_once_with(
            expected_cmd,
            cwd=openscap.tempfile.mkdtemp.return_value,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        openscap.__salt__["cp.push_dir"].assert_called_once_with(
            openscap.tempfile.mkdtemp.return_value
        )
        assert openscap.shutil.rmtree.call_count == 1
        assert response == {
            "upload_dir": openscap.tempfile.mkdtemp.return_value,
            "error": "",
            "success": True,
            "returncode": 0,
        }


def test_new_openscap_xccdf_eval_success_with_extra_ovalfiles(policy_file):
    with patch(
        "salt.modules.openscap.subprocess.Popen",
        MagicMock(
            return_value=Mock(**{"returncode": 0, "communicate.return_value": ("", "")})
        ),
    ):
        response = openscap.xccdf_eval(
            policy_file,
            ["/usr/share/xml/another-oval.xml", "/usr/share/xml/oval.xml"],
            profile="Default",
            oval_results=True,
            results="results.xml",
            report="report.html",
        )

        expected_cmd = [
            "oscap",
            "xccdf",
            "eval",
            "--oval-results",
            "--results",
            "results.xml",
            "--report",
            "report.html",
            "--profile",
            "Default",
            policy_file,
            "/usr/share/xml/another-oval.xml",
            "/usr/share/xml/oval.xml",
        ]
        openscap.subprocess.Popen.assert_called_once_with(
            expected_cmd,
            cwd=openscap.tempfile.mkdtemp.return_value,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        openscap.__salt__["cp.push_dir"].assert_called_once_with(
            openscap.tempfile.mkdtemp.return_value
        )
        assert openscap.shutil.rmtree.call_count == 1
        assert response == {
            "upload_dir": openscap.tempfile.mkdtemp.return_value,
            "error": "",
            "success": True,
            "returncode": 0,
        }


def test_new_openscap_xccdf_eval_success_with_failing_rules(policy_file):
    with patch(
        "salt.modules.openscap.subprocess.Popen",
        MagicMock(
            return_value=Mock(
                **{"returncode": 2, "communicate.return_value": ("", "some error")}
            )
        ),
    ):
        response = openscap.xccdf_eval(
            policy_file,
            profile="Default",
            oval_results=True,
            results="results.xml",
            report="report.html",
        )

        assert openscap.tempfile.mkdtemp.call_count == 1
        expected_cmd = [
            "oscap",
            "xccdf",
            "eval",
            "--oval-results",
            "--results",
            "results.xml",
            "--report",
            "report.html",
            "--profile",
            "Default",
            policy_file,
        ]
        openscap.subprocess.Popen.assert_called_once_with(
            expected_cmd,
            cwd=openscap.tempfile.mkdtemp.return_value,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )
        openscap.__salt__["cp.push_dir"].assert_called_once_with(
            openscap.tempfile.mkdtemp.return_value
        )
        assert openscap.shutil.rmtree.call_count == 1
        assert response == {
            "upload_dir": openscap.tempfile.mkdtemp.return_value,
            "error": "some error",
            "success": True,
            "returncode": 2,
        }


def test_new_openscap_xccdf_eval_evaluation_error(policy_file):
    with patch(
        "salt.modules.openscap.subprocess.Popen",
        MagicMock(
            return_value=Mock(
                **{
                    "returncode": 1,
                    "communicate.return_value": ("", "evaluation error"),
                }
            )
        ),
    ):
        response = openscap.xccdf_eval(
            policy_file,
            profile="Default",
            oval_results=True,
            results="results.xml",
            report="report.html",
        )

        assert response == {
            "upload_dir": None,
            "error": "evaluation error",
            "success": False,
            "returncode": 1,
        }



import pytest
import requests.models

import salt.utils.vault as vault
import salt.utils.vault.cache as vcache
import salt.utils.vault.client as vclient
import salt.utils.vault.kv as vkv
from tests.support.mock import MagicMock, Mock, patch


@pytest.fixture
def path():
    return "secret/some/path"


@pytest.fixture
def paths():
    return {
        "data": "secret/data/some/path",
        "metadata": "secret/metadata/some/path",
        "delete": "secret/data/some/path",
        "delete_versions": "secret/delete/some/path",
        "destroy": "secret/destroy/some/path",
    }


@pytest.fixture
def kvv1_meta_response():
    return {
        "request_id": "b82f2df7-a9b6-920c-0ed2-a3463b996f9e",
        "lease_id": "",
        "renewable": False,
        "lease_duration": 0,
        "data": {
            "accessor": "kv_f8731f1b",
            "config": {
                "default_lease_ttl": 0,
                "force_no_cache": False,
                "max_lease_ttl": 0,
            },
            "description": "key/value secret storage",
            "external_entropy_access": False,
            "local": False,
            "options": None,
            "path": "secret/",
            "seal_wrap": False,
            "type": "kv",
            "uuid": "1d9431ac-060a-9b63-4572-3ca7ffd78347",
        },
        "wrap_info": None,
        "warnings": None,
        "auth": None,
    }


@pytest.fixture
def kvv2_meta_response():
    return {
        "request_id": "b82f2df7-a9b6-920c-0ed2-a3463b996f9e",
        "lease_id": "",
        "renewable": False,
        "lease_duration": 0,
        "data": {
            "accessor": "kv_f8731f1b",
            "config": {
                "default_lease_ttl": 0,
                "force_no_cache": False,
                "max_lease_ttl": 0,
            },
            "description": "key/value secret storage",
            "external_entropy_access": False,
            "local": False,
            "options": {
                "version": "2",
            },
            "path": "secret/",
            "seal_wrap": False,
            "type": "kv",
            "uuid": "1d9431ac-060a-9b63-4572-3ca7ffd78347",
        },
        "wrap_info": None,
        "warnings": None,
        "auth": None,
    }


@pytest.fixture
def kvv1_info():
    return {
        "v2": False,
        "data": "secret/some/path",
        "metadata": "secret/some/path",
        "delete": "secret/some/path",
        "type": "kv",
    }


@pytest.fixture
def kvv2_info():
    return {
        "v2": True,
        "data": "secret/data/some/path",
        "metadata": "secret/metadata/some/path",
        "delete": "secret/data/some/path",
        "delete_versions": "secret/delete/some/path",
        "destroy": "secret/destroy/some/path",
        "type": "kv",
    }


@pytest.fixture
def no_kv_info():
    return {
        "v2": False,
        "data": "secret/some/path",
        "metadata": "secret/some/path",
        "delete": "secret/some/path",
        "type": None,
    }


@pytest.fixture
def kvv1_response():
    return {
        "request_id": "35df4df1-c3d8-b270-0682-ddb0160c7450",
        "lease_id": "",
        "renewable": False,
        "lease_duration": 0,
        "data": {
            "foo": "bar",
        },
        "wrap_info": None,
        "warnings": None,
        "auth": None,
    }


@pytest.fixture
def kvv2_response():
    return {
        "request_id": "35df4df1-c3d8-b270-0682-ddb0160c7450",
        "lease_id": "",
        "renewable": False,
        "lease_duration": 0,
        "data": {
            "data": {"foo": "bar"},
            "metadata": {
                "created_time": "2020-05-02T07:26:12.180848003Z",
                "deletion_time": "",
                "destroyed": False,
                "version": 1,
            },
        },
        "wrap_info": None,
        "warnings": None,
        "auth": None,
    }


@pytest.fixture
def kv_list_response():
    return {
        "request_id": "35df4df1-c3d8-b270-0682-ddb0160c7450",
        "lease_id": "",
        "renewable": False,
        "lease_duration": 0,
        "data": {
            "keys": ["foo"],
        },
        "wrap_info": None,
        "warnings": None,
        "auth": None,
    }


@pytest.fixture
def metadata_nocache():
    cache = Mock(spec=vcache.VaultCache)
    cache.get.return_value = None
    return cache


@pytest.fixture(params=["v1", "v2"])
def kv_meta(request, metadata_nocache):
    client = Mock(spec=vclient.AuthenticatedVaultClient)
    if request.param == "invalid":
        res = {"wrap_info": {}}
    else:
        res = request.getfixturevalue(f"kv{request.param}_meta_response")
    client.get.return_value = res
    return vkv.VaultKV(client, metadata_nocache)


@pytest.fixture(params=["v1", "v2"])
def kv_meta_cached(request):
    cache = Mock(spec=vcache.VaultCache)
    client = Mock(spec=vclient.AuthenticatedVaultClient)
    kv_meta_response = request.getfixturevalue(f"kv{request.param}_meta_response")
    client.get.return_value = kv_meta_response
    cache.get.return_value = {"secret/some/path": kv_meta_response["data"]}
    return vkv.VaultKV(client, cache)


@pytest.fixture
def kvv1(kvv1_info, kvv1_response, metadata_nocache, kv_list_response):
    client = Mock(spec=vclient.AuthenticatedVaultClient)
    client.get.return_value = kvv1_response
    client.post.return_value = True
    client.patch.side_effect = vclient.VaultPermissionDeniedError
    client.list.return_value = kv_list_response
    client.delete.return_value = True
    with patch("salt.utils.vault.kv.VaultKV.is_v2", Mock(return_value=kvv1_info)):
        yield vkv.VaultKV(client, metadata_nocache)


@pytest.fixture
def kvv2(kvv2_info, kvv2_response, metadata_nocache, kv_list_response):
    client = Mock(spec=vclient.AuthenticatedVaultClient)
    client.get.return_value = kvv2_response
    client.post.return_value = True
    client.patch.return_value = True
    client.list.return_value = kv_list_response
    client.delete.return_value = True
    with patch("salt.utils.vault.kv.VaultKV.is_v2", Mock(return_value=kvv2_info)):
        yield vkv.VaultKV(client, metadata_nocache)


@pytest.mark.parametrize(
    "wrapper,param,result",
    [
        ("read_kv", None, {"foo": "bar"}),
        ("write_kv", {"foo": "bar"}, True),
        ("patch_kv", {"foo": "bar"}, True),
        ("delete_kv", None, True),
        ("destroy_kv", [0], True),
        ("list_kv", None, ["foo"]),
    ],
)
@pytest.mark.parametrize("test_remote_config", ["token"], indirect=True)
@pytest.mark.parametrize(
    "clear_unauthd,token_valid", [(False, False), (True, False), (True, True)]
)
def test_kv_wrapper_handles_perm_exceptions(
    wrapper, param, result, test_remote_config, clear_unauthd, token_valid
):
    """
    Test that *_kv wrappers retry with a new client if
      a) the current configuration might be invalid
      b) the current token might not have all policies and
         `cache:clear_on_unauthorized` is True
    """
    func = getattr(vault, wrapper)
    exc = vault.VaultPermissionDeniedError
    args = ["secret/some/path"]
    if param:
        args.append(param)
    args += [{}, {}]
    test_remote_config["cache"]["clear_on_unauthorized"] = clear_unauthd
    with patch("salt.utils.vault.get_kv", autospec=True) as getkv:
        with patch("salt.utils.vault.clear_cache", autospec=True) as cache:
            kv = Mock(spec=vkv.VaultKV)
            kv.client = Mock(spec=vclient.AuthenticatedVaultClient)
            kv.client.token_valid.return_value = token_valid
            getattr(kv, wrapper.rstrip("_kv")).side_effect = (exc, result)
            getkv.side_effect = ((kv, test_remote_config), kv)
            res = func(*args)
            assert res == result
            cache.assert_called_once()


@pytest.mark.parametrize(
    "wrapper,param",
    [
        ("read_kv", None),
        ("write_kv", {"foo": "bar"}),
        ("patch_kv", {"foo": "bar"}),
        ("delete_kv", None),
        ("destroy_kv", [0]),
        ("list_kv", None),
    ],
)
@pytest.mark.parametrize("test_remote_config", ["token"], indirect=True)
def test_kv_wrapper_raises_perm_exceptions_when_configured(
    wrapper, param, test_remote_config
):
    """
    Test that *_kv wrappers do not retry with a new client when `cache:clear_on_unauthorized` is False.
    """
    func = getattr(vault, wrapper)
    exc = vault.VaultPermissionDeniedError
    args = ["secret/some/path"]
    if param:
        args.append(param)
    args += [{}, {}]
    test_remote_config["cache"]["clear_on_unauthorized"] = False
    with patch("salt.utils.vault.get_kv", autospec=True) as getkv:
        with patch("salt.utils.vault.clear_cache", autospec=True):
            kv = Mock(spec=vkv.VaultKV)
            kv.client = Mock(spec=vclient.AuthenticatedVaultClient)
            kv.client.token_valid.return_value = True
            getattr(kv, wrapper.rstrip("_kv")).side_effect = exc
            getkv.return_value = (kv, test_remote_config)
            with pytest.raises(exc):
                func(*args)


@pytest.mark.parametrize(
    "kv_meta,expected",
    [
        (
            "v1",
            "kvv1_info",
        ),
        (
            "v2",
            "kvv2_info",
        ),
        (
            "invalid",
            "no_kv_info",
        ),
    ],
    indirect=["kv_meta"],
)
def test_vault_kv_is_v2_no_cache(kv_meta, expected, request):
    """
    Ensure path metadata is requested as expected and cached
    if the lookup succeeds
    """
    expected_val = request.getfixturevalue(expected)
    res = kv_meta.is_v2("secret/some/path")
    kv_meta.metadata_cache.get.assert_called_once()
    kv_meta.client.get.assert_called_once_with(
        "sys/internal/ui/mounts/secret/some/path"
    )
    if expected != "no_kv_info":
        kv_meta.metadata_cache.store.assert_called_once()
    assert res == expected_val


@pytest.mark.parametrize(
    "kv_meta_cached,expected",
    [
        (
            "v1",
            "kvv1_info",
        ),
        (
            "v2",
            "kvv2_info",
        ),
    ],
    indirect=["kv_meta_cached"],
)
def test_vault_kv_is_v2_cached(kv_meta_cached, expected, request):
    """
    Ensure cache is respected for path metadata
    """
    expected = request.getfixturevalue(expected)
    res = kv_meta_cached.is_v2("secret/some/path")
    kv_meta_cached.metadata_cache.get.assert_called_once()
    kv_meta_cached.metadata_cache.store.assert_not_called()
    kv_meta_cached.client.assert_not_called()
    assert res == expected


class TestKVV1:
    path = "secret/some/path"

    @pytest.mark.parametrize("include_metadata", [False, True])
    def test_vault_kv_read(self, kvv1, include_metadata, path):
        """
        Ensure that VaultKV.read works for KV v1 and does not fail if
        metadata is requested, which is invalid for KV v1.
        """
        res = kvv1.read(path, include_metadata=include_metadata)
        kvv1.client.get.assert_called_once_with(path)
        assert res == {"foo": "bar"}

    def test_vault_kv_write(self, kvv1, path):
        """
        Ensure that VaultKV.write works for KV v1.
        """
        data = {"bar": "baz"}
        kvv1.write(path, data)
        kvv1.client.post.assert_called_once_with(path, payload=data)

    @pytest.mark.parametrize(
        "existing,data,expected",
        [
            ({"foo": "bar"}, {"bar": "baz"}, {"foo": "bar", "bar": "baz"}),
            ({"foo": "bar"}, {"foo": None}, {}),
            (
                {"foo": "bar"},
                {"foo2": {"bar": {"baz": True}}},
                {"foo": "bar", "foo2": {"bar": {"baz": True}}},
            ),
            (
                {"foo": {"bar": {"baz": True}}},
                {"foo": {"bar": {"baz": None}}},
                {"foo": {"bar": {}}},
            ),
        ],
    )
    def test_vault_kv_patch(self, kvv1, path, existing, data, expected):
        """
        Ensure that VaultKV.patch works for KV v1.
        This also tests the internal JSON merge patch implementation.
        """
        kvv1.client.get.return_value = {"data": existing}
        kvv1.patch(path, data)
        kvv1.client.post.assert_called_once_with(
            path,
            payload=expected,
        )

    def test_vault_kv_delete(self, kvv1, path):
        """
        Ensure that VaultKV.delete works for KV v1.
        """
        kvv1.delete(path)
        kvv1.client.request.assert_called_once_with("DELETE", path, payload=None)

    def test_vault_kv_delete_versions(self, kvv1, path):
        """
        Ensure that VaultKV.delete with versions raises an exception for KV v1.
        """
        with pytest.raises(
            vault.VaultInvocationError, match="Versioning support requires kv-v2.*"
        ):
            kvv1.delete(path, versions=[1, 2, 3, 4])

    def test_vault_kv_destroy(self, kvv1, path):
        """
        Ensure that VaultKV.destroy raises an exception for KV v1.
        """
        with pytest.raises(vault.VaultInvocationError):
            kvv1.destroy(path, [1, 2, 3, 4])

    def test_vault_kv_nuke(self, kvv1, path):
        """
        Ensure that VaultKV.nuke raises an exception for KV v1.
        """
        with pytest.raises(vault.VaultInvocationError):
            kvv1.nuke(path)

    def test_vault_kv_list(self, kvv1, path):
        """
        Ensure that VaultKV.list works for KV v1 and only returns keys.
        """
        res = kvv1.list(path)
        kvv1.client.list.assert_called_once_with(path)
        assert res == ["foo"]


class TestKVV2:
    @pytest.mark.parametrize(
        "versions,expected",
        [
            (0, [0]),
            ("1", [1]),
            ([2], [2]),
            (["3"], [3]),
        ],
    )
    def test_parse_versions(self, kvv2, versions, expected):
        """
        Ensure parsing versions works as expected:
        single integer/number string or list of those are allowed
        """
        assert kvv2._parse_versions(versions) == expected

    def test_parse_versions_raises_exception_when_unparsable(self, kvv2):
        """
        Ensure unparsable versions raise an exception
        """
        with pytest.raises(vault.VaultInvocationError):
            kvv2._parse_versions("four")

    def test_get_secret_path_metadata_lookup_unexpected_response(
        self, kvv2, caplog, path
    ):
        """
        Ensure unexpected responses are treated as not KV
        """
        # _mock_json_response() returns a Mock, but we need MagicMock here
        resp_mm = MagicMock(spec=requests.models.Response)
        resp_mm.json.return_value = {"wrap_info": {}}
        resp_mm.status_code = 200
        resp_mm.reason = ""
        kvv2.client.get.return_value = resp_mm
        res = kvv2._get_secret_path_metadata(path)
        assert res is None
        assert "Unexpected response to metadata query" in caplog.text

    def test_get_secret_path_metadata_lookup_request_error(self, kvv2, caplog, path):
        """
        Ensure HTTP error status codes are treated as not KV
        """
        kvv2.client.get.side_effect = vault.VaultPermissionDeniedError
        res = kvv2._get_secret_path_metadata(path)
        assert res is None
        assert "VaultPermissionDeniedError:" in caplog.text

    @pytest.mark.parametrize("include_metadata", [False, True])
    def test_vault_kv_read(self, kvv2, include_metadata, kvv2_response, paths):
        """
        Ensure that VaultKV.read works for KV v2 and returns metadata
        if requested.
        """
        res = kvv2.read(path, include_metadata=include_metadata)
        kvv2.client.get.assert_called_once_with(paths["data"])
        if include_metadata:
            assert res == kvv2_response["data"]
        else:
            assert res == kvv2_response["data"]["data"]

    def test_vault_kv_write(self, kvv2, path, paths):
        """
        Ensure that VaultKV.write works for KV v2.
        """
        data = {"bar": "baz"}
        kvv2.write(path, data)
        kvv2.client.post.assert_called_once_with(paths["data"], payload={"data": data})

    def test_vault_kv_patch(self, kvv2, path, paths):
        """
        Ensure that VaultKV.patch works for KV v2.
        """
        data = {"bar": "baz"}
        kvv2.patch(path, data)
        kvv2.client.patch.assert_called_once_with(
            paths["data"],
            payload={"data": data},
            add_headers={"Content-Type": "application/merge-patch+json"},
        )

    def test_vault_kv_delete(self, kvv2, path, paths):
        """
        Ensure that VaultKV.delete works for KV v2.
        """
        kvv2.delete(path)
        kvv2.client.request.assert_called_once_with(
            "DELETE", paths["data"], payload=None
        )

    @pytest.mark.parametrize(
        "versions", [[1, 2], [2], 2, ["1", "2"], ["2"], "2", [1, "2"]]
    )
    def test_vault_kv_delete_versions(self, kvv2, versions, path, paths):
        """
        Ensure that VaultKV.delete with versions works for KV v2.
        """
        if isinstance(versions, list):
            expected = [int(x) for x in versions]
        else:
            expected = [int(versions)]
        kvv2.delete(path, versions=versions)
        kvv2.client.request.assert_called_once_with(
            "POST", paths["delete_versions"], payload={"versions": expected}
        )

    @pytest.mark.parametrize(
        "versions", [[1, 2], [2], 2, ["1", "2"], ["2"], "2", [1, "2"]]
    )
    def test_vault_kv_destroy(self, kvv2, versions, path, paths):
        """
        Ensure that VaultKV.destroy works for KV v2.
        """
        if isinstance(versions, list):
            expected = [int(x) for x in versions]
        else:
            expected = [int(versions)]
        kvv2.destroy(path, versions)
        kvv2.client.post.assert_called_once_with(
            paths["destroy"], payload={"versions": expected}
        )

    def test_vault_kv_nuke(self, kvv2, path, paths):
        """
        Ensure that VaultKV.nuke works for KV v2.
        """
        kvv2.nuke(path)
        kvv2.client.delete.assert_called_once_with(paths["metadata"])

    def test_vault_kv_list(self, kvv2, path, paths):
        """
        Ensure that VaultKV.list works for KV v2 and only returns keys.
        """
        res = kvv2.list(path)
        kvv2.client.list.assert_called_once_with(paths["metadata"])
        assert res == ["foo"]



import logging
import os
import shutil
import textwrap

import pytest

import salt.config
import salt.loader
import salt.modules.cmdmod as cmdmod
import salt.modules.file as filemod
import salt.utils.data
import salt.utils.files
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

if salt.utils.platform.is_windows():
    import salt.modules.win_file as win_file
    import salt.utils.win_dacl as win_dacl

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    if salt.utils.platform.is_windows():
        grains = {"kernel": "Windows"}
    else:
        grains = {"kernel": "Linux"}
    opts = {
        "test": False,
        "file_roots": {"base": "tmp"},
        "pillar_roots": {"base": "tmp"},
        "cachedir": "tmp",
        "grains": grains,
    }

    ret = {
        filemod: {
            "__salt__": {
                "config.manage_mode": MagicMock(),
                "cmd.run": cmdmod.run,
                "cmd.run_all": cmdmod.run_all,
            },
            "__opts__": opts,
            "__grains__": grains,
            "__utils__": {
                "files.is_binary": MagicMock(return_value=False),
                "files.is_text": salt.utils.files.is_text,
                "files.get_encoding": MagicMock(return_value="utf-8"),
                "stringutils.get_diff": salt.utils.stringutils.get_diff,
            },
        }
    }
    if salt.utils.platform.is_windows():
        ret.update(
            {
                win_dacl: {"__opts__": opts},
                win_file: {
                    "__utils__": {"dacl.check_perms": win_dacl.check_perms},
                    "__opts__": opts,
                },
            }
        )

    return ret


@pytest.fixture
def multiline_string():
    multiline_string = textwrap.dedent(
        """\
        Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam rhoncus
        enim ac bibendum vulputate. Etiam nibh velit, placerat ac auctor in,
        lacinia a turpis. Nulla elit elit, ornare in sodales eu, aliquam sit
        amet nisl.

        Fusce ac vehicula lectus. Vivamus justo nunc, pulvinar in ornare nec,
        sollicitudin id sem. Pellentesque sed ipsum dapibus, dapibus elit id,
        malesuada nisi.

        first part of start line // START BLOCK : part of start line not removed
        to be removed
        first part of end line // END BLOCK : part of end line not removed

        #-- START BLOCK UNFINISHED

        #-- START BLOCK 1
        old content part 1
        old content part 2
        #-- END BLOCK 1

        Lorem ipsum dolor sit amet, consectetur adipiscing elit. Donec
        venenatis tellus eget massa facilisis, in auctor ante aliquet. Sed nec
        cursus metus. Curabitur massa urna, vehicula id porttitor sed, lobortis
        quis leo.
        """
    )

    multiline_string = os.linesep.join(multiline_string.splitlines())

    return multiline_string


@pytest.fixture
def multiline_file(tmp_path, multiline_string):
    multiline_file = str(tmp_path / "multiline-file.txt")

    with salt.utils.files.fopen(multiline_file, "w+b") as file_handle:
        file_handle.write(salt.utils.stringutils.to_bytes(multiline_string))

    yield multiline_file
    shutil.rmtree(str(tmp_path))


# Make a unique subdir to avoid any tempfile conflicts
@pytest.fixture
def subdir(tmp_path):
    subdir = tmp_path / "test-file-block-replace-subdir"
    subdir.mkdir()
    yield subdir
    shutil.rmtree(str(subdir))


def test_replace_multiline(multiline_file):
    new_multiline_content = os.linesep.join(
        [
            "Who's that then?",
            "Well, how'd you become king, then?",
            "We found them. I'm not a witch.",
            "We shall say 'Ni' again to you, if you do not appease us.",
        ]
    )
    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 1",
            marker_end="#-- END BLOCK 1",
            content=new_multiline_content,
            backup=False,
            append_newline=None,
        )

    with salt.utils.files.fopen(multiline_file, "rb") as fp:
        filecontent = fp.read()

    assert (
        salt.utils.stringutils.to_bytes(
            os.linesep.join(
                ["#-- START BLOCK 1", new_multiline_content, "#-- END BLOCK 1"]
            )
        )
        in filecontent
    )
    assert b"old content part 1" not in filecontent
    assert b"old content part 2" not in filecontent


def test_replace_append(multiline_file):
    new_content = "Well, I didn't vote for you."

    with pytest.raises(CommandExecutionError):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 2",
            marker_end="#-- END BLOCK 2",
            content=new_content,
            append_if_not_found=False,
            backup=False,
        )

    with pytest.raises(CommandExecutionError):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 2",
            marker_end="#-- END BLOCK 2",
            content=new_content,
            append_if_not_found=False,
            backup=False,
        )
    with salt.utils.files.fopen(multiline_file, "r") as fp:
        assert (
            "#-- START BLOCK 2" + "\n" + new_content + "#-- END BLOCK 2"
            not in salt.utils.stringutils.to_unicode(fp.read())
        )

    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 2",
            marker_end="#-- END BLOCK 2",
            content=new_content,
            backup=False,
            append_if_not_found=True,
        )

    with salt.utils.files.fopen(multiline_file, "rb") as fp:
        assert (
            salt.utils.stringutils.to_bytes(
                os.linesep.join(["#-- START BLOCK 2", f"{new_content}#-- END BLOCK 2"])
            )
            in fp.read()
        )


def test_replace_insert_after(multiline_file):
    new_content = "Well, I didn't vote for you."

    with pytest.raises(CommandExecutionError):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 2",
            marker_end="#-- END BLOCK 2",
            content=new_content,
            insert_after_match="not in the text",
            backup=False,
        )
    with salt.utils.files.fopen(multiline_file, "r") as fp:
        assert (
            "#-- START BLOCK 2" + "\n" + new_content + "#-- END BLOCK 2"
            not in salt.utils.stringutils.to_unicode(fp.read())
        )

    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 2",
            marker_end="#-- END BLOCK 2",
            content=new_content,
            backup=False,
            insert_after_match="malesuada",
        )

    with salt.utils.files.fopen(multiline_file, "rb") as fp:
        assert (
            salt.utils.stringutils.to_bytes(
                os.linesep.join(["#-- START BLOCK 2", f"{new_content}#-- END BLOCK 2"])
            )
            in fp.read()
        )


def test_replace_append_newline_at_eof(subdir):
    """
    Check that file.blockreplace works consistently on files with and
    without newlines at end of file.
    """
    base = "bar"
    args = {
        "marker_start": "#start",
        "marker_end": "#stop",
        "content": "baz",
        "append_if_not_found": True,
    }
    block = os.linesep.join(["#start", "baz#stop"]) + os.linesep
    # File ending with a newline
    with salt.utils.files.fopen(str(subdir / "tfile"), "w+b") as tfile:
        tfile.write(salt.utils.stringutils.to_bytes(base + os.linesep))
        tfile.flush()
    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(tfile.name, **args)
    expected = os.linesep.join([base, block])
    with salt.utils.files.fopen(tfile.name) as tfile2:
        assert salt.utils.stringutils.to_unicode(tfile2.read()) == expected
    os.remove(tfile.name)

    # File not ending with a newline
    with salt.utils.files.fopen(str(subdir / "tfile"), "w+b") as tfile:
        tfile.write(salt.utils.stringutils.to_bytes(base))
        tfile.flush()
    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(tfile.name, **args)
    with salt.utils.files.fopen(tfile.name) as tfile2:
        assert salt.utils.stringutils.to_unicode(tfile2.read()) == expected
    os.remove(tfile.name)

    # A newline should not be added in empty files
    with salt.utils.files.fopen(str(subdir / "tfile"), "w+b") as tfile:
        pass
    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(tfile.name, **args)
    with salt.utils.files.fopen(tfile.name) as tfile2:
        assert salt.utils.stringutils.to_unicode(tfile2.read()) == block
    os.remove(tfile.name)


def test_replace_prepend(multiline_file):
    new_content = "Well, I didn't vote for you."

    with pytest.raises(CommandExecutionError):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 2",
            marker_end="#-- END BLOCK 2",
            content=new_content,
            prepend_if_not_found=False,
            backup=False,
        )
    with salt.utils.files.fopen(multiline_file, "rb") as fp:
        assert (
            salt.utils.stringutils.to_bytes(
                os.linesep.join(["#-- START BLOCK 2", f"{new_content}#-- END BLOCK 2"])
            )
            not in fp.read()
        )

    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 2",
            marker_end="#-- END BLOCK 2",
            content=new_content,
            backup=False,
            prepend_if_not_found=True,
        )

    with salt.utils.files.fopen(multiline_file, "rb") as fp:
        assert fp.read().startswith(
            salt.utils.stringutils.to_bytes(
                os.linesep.join(
                    [
                        "#-- START BLOCK 2",
                        f"{new_content}#-- END BLOCK 2",
                    ]
                )
            )
        )


def test_replace_insert_before(multiline_file):
    new_content = "Well, I didn't vote for you."

    with pytest.raises(CommandExecutionError):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 2",
            marker_end="#-- END BLOCK 2",
            content=new_content,
            insert_before_match="not in the text",
            backup=False,
        )
    with salt.utils.files.fopen(multiline_file, "r") as fp:
        assert (
            "#-- START BLOCK 2" + "\n" + new_content + "#-- END BLOCK 2"
            not in salt.utils.stringutils.to_unicode(fp.read())
        )

    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 2",
            marker_end="#-- END BLOCK 2",
            content=new_content,
            backup=False,
            insert_before_match="malesuada",
        )

    with salt.utils.files.fopen(multiline_file, "rb") as fp:
        assert (
            salt.utils.stringutils.to_bytes(
                os.linesep.join(["#-- START BLOCK 2", f"{new_content}#-- END BLOCK 2"])
            )
            in fp.read()
        )


def test_replace_partial_marked_lines(multiline_file):
    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(
            multiline_file,
            marker_start="// START BLOCK",
            marker_end="// END BLOCK",
            content="new content 1",
            backup=False,
        )

    with salt.utils.files.fopen(multiline_file, "r") as fp:
        filecontent = salt.utils.stringutils.to_unicode(fp.read())
    assert "new content 1" in filecontent
    assert "to be removed" not in filecontent
    assert "first part of start line" in filecontent
    assert "first part of end line" not in filecontent
    assert "part of start line not removed" in filecontent
    assert "part of end line not removed" in filecontent


def test_backup(multiline_file):
    fext = ".bak"
    bak_file = f"{multiline_file}{fext}"

    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(
            multiline_file,
            marker_start="// START BLOCK",
            marker_end="// END BLOCK",
            content="new content 2",
            backup=fext,
        )

    assert os.path.exists(bak_file)
    os.unlink(bak_file)
    assert not os.path.exists(bak_file)

    fext = ".bak"
    bak_file = f"{multiline_file}{fext}"

    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(
            multiline_file,
            marker_start="// START BLOCK",
            marker_end="// END BLOCK",
            content="new content 3",
            backup=False,
        )

    assert not os.path.exists(bak_file)


def test_no_modifications(multiline_file):
    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 1",
            marker_end="#-- END BLOCK 1",
            content="new content 4",
            backup=False,
            append_newline=None,
        )
    before_ctime = os.stat(multiline_file).st_mtime
    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK 1",
            marker_end="#-- END BLOCK 1",
            content="new content 4",
            backup=False,
            append_newline=None,
        )
    after_ctime = os.stat(multiline_file).st_mtime

    assert before_ctime == after_ctime


def test_dry_run(multiline_file):
    before_ctime = os.stat(multiline_file).st_mtime
    filemod.blockreplace(
        multiline_file,
        marker_start="// START BLOCK",
        marker_end="// END BLOCK",
        content="new content 5",
        dry_run=True,
    )
    after_ctime = os.stat(multiline_file).st_mtime

    assert before_ctime == after_ctime


def test_show_changes(multiline_file):
    if salt.utils.platform.is_windows():
        check_perms_patch = win_file.check_perms
    else:
        check_perms_patch = filemod.check_perms
    with patch.object(filemod, "check_perms", check_perms_patch):
        ret = filemod.blockreplace(
            multiline_file,
            marker_start="// START BLOCK",
            marker_end="// END BLOCK",
            content="new content 6",
            backup=False,
            show_changes=True,
        )

        assert ret.startswith("---")  # looks like a diff

        ret = filemod.blockreplace(
            multiline_file,
            marker_start="// START BLOCK",
            marker_end="// END BLOCK",
            content="new content 7",
            backup=False,
            show_changes=False,
        )

        assert isinstance(ret, bool)


def test_unfinished_block_exception(multiline_file):
    with pytest.raises(CommandExecutionError):
        filemod.blockreplace(
            multiline_file,
            marker_start="#-- START BLOCK UNFINISHED",
            marker_end="#-- END BLOCK UNFINISHED",
            content="foobar",
            backup=False,
        )


def test_search_proc_file():
    """
    Test that searching content in a /proc file does not raise a TypeError
    and handles bytes correctly.
    """
    proc_file_path = "/proc/cpuinfo"

    if not os.path.exists(proc_file_path):
        pytest.skip(f"{proc_file_path} not available")

    match_found = filemod.search(proc_file_path, "processor")
    assert match_found, "Failed to find 'processor' in /proc/cpuinfo"



import time

import salt.auth
import salt.config


def test_cve_2021_3244(tmp_path):
    token_dir = tmp_path / "tokens"
    token_dir.mkdir()
    opts = {
        "extension_modules": "",
        "optimization_order": [0, 1, 2],
        "token_expire": 1,
        "keep_acl_in_token": False,
        "eauth_tokens": "localfs",
        "token_dir": str(token_dir),
        "token_expire_user_override": True,
        "external_auth": {"auto": {"foo": []}},
    }
    auth = salt.auth.LoadAuth(opts)
    load = {
        "eauth": "auto",
        "username": "foo",
        "password": "foo",
        "token_expire": -1,
    }
    t_data = auth.mk_token(load)
    assert t_data["expire"] < time.time()
    token_file = token_dir / t_data["token"]
    assert token_file.exists()
    t_data = auth.get_tok(t_data["token"])
    assert not t_data
    assert not token_file.exists()



"""
Utilities for working with etcd

.. versionadded:: 2014.7.0

:depends:  - python-etcd or etcd3-py

This library sets up a client object for etcd, using the configuration passed
into the get_conn() function. Normally, this is __opts__. Optionally, a profile
may be passed in. The following configurations are both valid:

.. code-block:: yaml

    # No profile name
    etcd.host: 127.0.0.1
    etcd.port: 2379
    etcd.username: larry  # Optional; requires etcd.password to be set
    etcd.password: 123pass  # Optional; requires etcd.username to be set
    etcd.ca: /path/to/your/ca_cert/ca.pem # Optional
    etcd.client_key: /path/to/your/client_key/client-key.pem # Optional; requires etcd.ca and etcd.client_cert to be set
    etcd.client_cert: /path/to/your/client_cert/client.pem # Optional; requires etcd.ca and etcd.client_key to be set
    etcd.require_v2: True # Optional; defaults to True
    etcd.encode_keys: False # Optional (v3 ONLY); defaults to False
    etcd.encode_values: True # Optional (v3 ONLY); defaults to True
    etcd.raw_keys: False # Optional (v3 ONLY); defaults to False
    etcd.raw_values: False # Optional (v3 ONLY); defaults to False
    etcd.unicode_errors: "surrogateescape" # Optional (v3 ONLY); defaults to "surrogateescape"

    # One or more profiles defined
    my_etcd_config:
      etcd.host: 127.0.0.1
      etcd.port: 2379
      etcd.username: larry  # Optional; requires etcd.password to be set
      etcd.password: 123pass  # Optional; requires etcd.username to be set
      etcd.ca: /path/to/your/ca_cert/ca.pem # Optional
      etcd.client_key: /path/to/your/client_key/client-key.pem # Optional; requires etcd.ca and etcd.client_cert to be set
      etcd.client_cert: /path/to/your/client_cert/client.pem # Optional; requires etcd.ca and etcd.client_key to be set
      etcd.require_v2: True # Optional; defaults to True
      etcd.encode_keys: False # Optional (v3 ONLY); defaults to False
      etcd.encode_values: True # Optional (v3 ONLY); defaults to True
      etcd.raw_keys: False # Optional (v3 ONLY); defaults to False
      etcd.raw_values: False # Optional (v3 ONLY); defaults to False
      etcd.unicode_errors: "surrogateescape" # Optional (v3 ONLY); defaults to "surrogateescape"

Encoding keys for etcd v3 allows a differentiation within etcd between byte and string keys.
It is worth noting that if you chose to encode keys, due to the way encoding pre-etcd with msgpack works,
all recursive functionality will not work as intended. This includes tree and ls along
with all methods that have recurse kwargs.  Thus, enabling this option is not recommended.

Once configured, the client() function is passed a set of opts, and optionally,
the name of a profile to be used.

.. code-block:: python

    import salt.utils.etcd_utils
    client = salt.utils.etcd_utils.get_conn(__opts__, profile='my_etcd_config')

You may also use the newer syntax and bypass the generator function.

V2 API
.. code-block:: python

    import salt.utils.etcd_utils
    client = salt.utils.etcd_utils.EtcdClient(__opts__, profile='my_etcd_config')

V3 API
.. versionadded:: 3005

.. code-block:: python

    import salt.utils.etcd_utils
    client = salt.utils.etcd_utils.EtcdClientV3(__opts__, profile='my_etcd_config')

It should be noted that some usages of etcd require a profile to be specified,
rather than top-level configurations. This being the case, it is better to
always use a named configuration profile, as shown above.
"""

import logging

import salt.utils.msgpack
import salt.utils.versions
from salt.exceptions import SaltException

try:
    import etcd
    from urllib3.exceptions import MaxRetryError, ReadTimeoutError

    HAS_ETCD_V2 = True
except ImportError:
    HAS_ETCD_V2 = False

try:
    import etcd3

    HAS_ETCD_V3 = True
except ImportError:
    HAS_ETCD_V3 = False

# Set up logging
log = logging.getLogger(__name__)


class EtcdLibraryNotInstalled(SaltException):
    """
    We didn't find the required etcd library
    """


class IncompatibleEtcdRequirements(SaltException):
    """
    A user is explicitly creating a client class, but requires a different version
    """


class Etcd3DirectoryException(SaltException):
    """
    We didn't find the required etcd library
    """


class EtcdUtilWatchTimeout(Exception):
    """
    A watch timed out without returning a result
    """


class EtcdBase:
    """
    Base class for the different versions of etcd clients.

    This also serves as a documentation hub for all superclasses.
    """

    def __init__(
        self,
        opts,
        profile=None,
        host=None,
        port=None,
        username=None,
        password=None,
        ca=None,
        client_key=None,
        client_cert=None,
        **kwargs,
    ):
        if not kwargs.get("has_etcd_opts", False):
            etcd_opts = _get_etcd_opts(opts, profile)
        else:
            etcd_opts = opts

        self.conf = etcd_opts
        self.host = host or self.conf.get("etcd.host", "127.0.0.1")
        self.port = port or self.conf.get("etcd.port", 2379)

        if self.host == "127.0.0.1" and self.port == 2379:
            log.warning("Using default etcd host and port, use a profile if needed.")

        username = username or self.conf.get("etcd.username")
        password = password or self.conf.get("etcd.password")
        ca_cert = ca or self.conf.get("etcd.ca")
        cli_key = client_key or self.conf.get("etcd.client_key")
        cli_cert = client_cert or self.conf.get("etcd.client_cert")

        auth = {}
        if username and password:
            auth = {
                "username": str(username),
                "password": str(password),
            }

        certs = {}
        if ca_cert and not (cli_cert or cli_key):
            certs = {"ca_cert": str(ca_cert), "protocol": "https"}

        if ca_cert and cli_cert and cli_key:
            cert = (cli_cert, cli_key)
            certs = {
                "ca_cert": str(ca_cert),
                "cert": cert,
                "protocol": "https",
            }

        self.xargs = auth.copy()
        self.xargs.update(certs)

    def watch(self, key, recurse=False, timeout=0, start_revision=None, **kwargs):
        raise NotImplementedError()

    def get(self, key, recurse=False):
        """
        Get the value of a specific key.  If recurse is true, defer to EtcdBase.tree() instead.
        """
        raise NotImplementedError()

    def read(
        self,
        key,
        recurse=False,
        wait=False,
        timeout=None,
        start_revision=None,
        **kwargs,
    ):
        """
        Read a value of a key.

        This method also provides the ability to wait for changes after a given index and/or
        within a certain timeout.
        """
        raise NotImplementedError()

    def _flatten(self, data, path=""):
        """
        Take a data dictionary and flatten it to a dictionary with values that are all strings.

        If path is given, prepend it to all keys.

        For example, given path="/salt" it will convert...

        {
            "key1": "value1",
            "key2": {
                "subkey1": "subvalue1",
                "subkey2": "subvalue2",
            }
        }

        to...

        {
            "/salt/key1": "value1",
            "/salt/key2/subkey1": "subvalue1",
            "/salt/key2/subkey2": "subvalue2",
        }
        """
        if not data:
            return {path: {}}
        path = path.strip("/")
        flat = {}
        for k, v in data.items():
            k = k.strip("/")
            if path:
                p = f"/{path}/{k}"
            else:
                p = f"/{k}"
            if isinstance(v, dict):
                ret = self._flatten(v, p)
                flat.update(ret)
            else:
                flat[p] = v
        return flat

    def update(self, fields, path=""):
        """
        Update etcd according to the layout of fields.

        Given etcd with this layout...
        {
            ...
            "/salt/key1": "OLDvalue1",
            "/salt/key2/subkey1": "OLDsubvalue1",
            "/salt/key2/subkey2": "OLDsubvalue2",
            ...
        }

        fields = {
            "key1": "value1",
            "key2": {
                "subkey1": "subvalue1",
                "subkey2": "subvalue2",
            }
        }

        will update etcd to look like the following...
        {
            ...
            "/salt/key1": "value1",
            "/salt/key2/subkey1": "subvalue1",
            "/salt/key2/subkey2": "subvalue2",
            ...
        }

        """
        raise NotImplementedError()

    def set(self, key, value, ttl=None, directory=False):
        """
        Write a file or directory, a higher interface to write
        """
        return self.write(key, value, ttl=ttl, directory=directory)

    def write(self, key, value, ttl=None, directory=False):
        """
        Write a file or directory depending on directory flag
        """
        if directory:
            return self.write_directory(key, value, ttl)
        return self.write_file(key, value, ttl)

    def write_file(self, key, value, ttl=None):
        """
        Write a file (key: value pair) to etcd
        """
        raise NotImplementedError()

    def write_directory(self, key, value, ttl=None):
        """
        Write a directory (key: {}) to etcd
        """
        raise NotImplementedError()

    def ls(self, path):
        """
        Get all the top level keys and their values at the given path.

        If the key is a directory, its value is an empty dictionary.
        """
        raise NotImplementedError()

    def rm(self, key, recurse=False):
        """
        An alias for delete
        """
        return self.delete(key, recurse)

    def delete(self, key, recurse=False, **kwargs):
        """
        Delete keys or (recursively) whole directories
        """
        raise NotImplementedError()

    def tree(self, path):
        """
        .. versionadded:: 2014.7.0

        Recurse through etcd and return all values
        """
        raise NotImplementedError()


class EtcdClient(EtcdBase):
    def __init__(self, opts, **kwargs):
        if not HAS_ETCD_V2:
            raise EtcdLibraryNotInstalled("Don't have python-etcd, need to install it.")
        log.debug("etcd_util has the libraries needed for etcd v2")

        super().__init__(opts, **kwargs)

        if not self.conf.get("etcd.require_v2", True):
            raise IncompatibleEtcdRequirements("Can't create v2 with a v3 requirement")

        self.client = etcd.Client(host=self.host, port=self.port, **self.xargs)

    def watch(self, key, recurse=False, timeout=0, start_revision=None, **kwargs):
        index = kwargs.pop("index", None)
        if index is not None:
            salt.utils.versions.warn_until(
                3008,
                "The index kwarg has been deprecated, and will be removed "
                "in the Argon release. Please use start_revision instead.",
            )
            start_revision = index
        if kwargs:
            log.warning("Invalid kwargs passed in will not be used: %s", kwargs)

        ret = {"key": key, "value": None, "changed": False, "mIndex": 0, "dir": False}
        try:
            result = self.read(
                key,
                recurse=recurse,
                wait=True,
                timeout=timeout,
                start_revision=start_revision,
            )
        except EtcdUtilWatchTimeout:
            try:
                result = self.read(key)
            except etcd.EtcdKeyNotFound:
                log.debug("etcd: key was not created while watching")
                return ret
            except ValueError:
                return {}
            if result and getattr(result, "dir"):
                ret["dir"] = True
            ret["value"] = getattr(result, "value")
            ret["mIndex"] = getattr(result, "modifiedIndex")
            return ret
        except MaxRetryError:
            # This gets raised when we can't contact etcd at all
            log.error(
                "etcd: failed to perform 'watch' operation on key %s due to connection"
                " error",
                key,
            )
            return {}
        except etcd.EtcdConnectionFailed as err:
            log.error("etcd: %s", err)
            return None
        except ValueError:
            return {}

        if result is None:
            return {}

        if recurse:
            ret["key"] = getattr(result, "key", None)
        ret["value"] = getattr(result, "value", None)
        ret["dir"] = getattr(result, "dir", None)
        ret["changed"] = True
        ret["mIndex"] = getattr(result, "modifiedIndex")
        return ret

    def get(self, key, recurse=False):
        if not recurse:
            try:
                result = self.read(key)
            except etcd.EtcdKeyNotFound:
                # etcd already logged that the key wasn't found, no need to do
                # anything here but return
                return None
            except etcd.EtcdConnectionFailed:
                log.error(
                    "etcd: failed to perform 'get' operation on key %s due to connection"
                    " error",
                    key,
                )
                return None
            except ValueError:
                return None

            return getattr(result, "value", None)

        return self.tree(key)

    def read(
        self,
        key,
        recurse=False,
        wait=False,
        timeout=None,
        start_revision=None,
        **kwargs,
    ):
        recursive = kwargs.pop("recursive", None)
        wait_index = kwargs.pop("waitIndex", None)
        if recursive is not None:
            salt.utils.versions.warn_until(
                3008,
                "The recursive kwarg has been deprecated, and will be removed "
                "in the Argon release. Please use recurse instead.",
            )
            recurse = recursive
        if wait_index is not None:
            salt.utils.versions.warn_until(
                3008,
                "The waitIndex kwarg has been deprecated, and will be removed "
                "in the Argon release. Please use start_revision instead.",
            )
            start_revision = wait_index
        if kwargs:
            log.warning("Invalid kwargs passed in will not be used: %s", kwargs)

        try:
            if start_revision:
                result = self.client.read(
                    key,
                    recursive=recurse,
                    wait=wait,
                    timeout=timeout,
                    waitIndex=start_revision,
                )
            else:
                result = self.client.read(
                    key, recursive=recurse, wait=wait, timeout=timeout
                )
        except (etcd.EtcdConnectionFailed, etcd.EtcdKeyNotFound) as err:
            log.error("etcd: %s", err)
            raise
        except ReadTimeoutError:
            # For some reason, we have to catch this directly.  It falls through
            # from python-etcd because it's trying to catch
            # urllib3.exceptions.ReadTimeoutError and strangely, doesn't catch.
            # This can occur from a watch timeout that expires, so it may be 'expected'
            # behavior. See issue #28553
            if wait:
                # Wait timeouts will throw ReadTimeoutError, which isn't bad
                log.debug("etcd: Timed out while executing a wait")
                raise EtcdUtilWatchTimeout(f"Watch on {key} timed out")
            log.error("etcd: Timed out")
            raise etcd.EtcdConnectionFailed("Connection failed")
        except MaxRetryError as err:
            # Same issue as ReadTimeoutError.  When it 'works', python-etcd
            # throws EtcdConnectionFailed, so we'll do that for it.
            log.error("etcd: Could not connect")
            raise etcd.EtcdConnectionFailed("Could not connect to etcd server")
        except etcd.EtcdException as err:
            # EtcdValueError inherits from ValueError, so we don't want to accidentally
            # catch this below on ValueError and give a bogus error message
            log.error("etcd: %s", err)
            raise
        except ValueError:
            # python-etcd doesn't fully support python 2.6 and ends up throwing this for *any* exception because
            # it uses the newer {} format syntax
            log.error(
                "etcd: error. python-etcd does not fully support python 2.6, no error"
                " information available"
            )
            raise
        except Exception as err:  # pylint: disable=broad-except
            log.error("etcd: uncaught exception %s", err)
            raise
        return result

    def update(self, fields, path=""):
        if not isinstance(fields, dict):
            log.error("etcd.update: fields is not type dict")
            return None
        fields = self._flatten(fields, path)
        keys = {}
        for k, v in fields.items():
            is_dir = False
            if isinstance(v, dict):
                is_dir = True
            keys[k] = self.write(k, v, directory=is_dir)
            if keys[k] is None:
                return None
        return keys

    def write(self, key, value, ttl=None, directory=False):
        """
        Write a file or directory depending on directory flag
        """
        try:
            if directory:
                return self.write_directory(key, value, ttl)
            return self.write_file(key, value, ttl)
        except etcd.EtcdConnectionFailed as err:
            log.error("etcd: %s", err)
            return None

    def write_file(self, key, value, ttl=None):
        try:
            result = self.client.write(key, value, ttl=ttl, dir=False)
        except (etcd.EtcdNotFile, etcd.EtcdRootReadOnly, ValueError) as err:
            # If EtcdNotFile is raised, then this key is a directory and
            # really this is a name collision.
            log.error("etcd: %s", err)
            return None
        except MaxRetryError as err:
            log.error("etcd: Could not connect to etcd server: %s", err)
            return None
        except Exception as err:  # pylint: disable=broad-except
            log.error("etcd: uncaught exception %s", err)
            raise

        return getattr(result, "value")

    def write_directory(self, key, value, ttl=None):
        if value is not None:
            log.info("etcd: non-empty value passed for directory: %s", value)
        try:
            # directories can't have values, but have to have it passed
            result = self.client.write(key, None, ttl=ttl, dir=True)
        except etcd.EtcdNotFile:
            # When a directory already exists, python-etcd raises an EtcdNotFile
            # exception. In this case, we just catch and return True for success.
            log.info("etcd: directory already exists: %s", key)
            return True
        except (etcd.EtcdNotDir, etcd.EtcdRootReadOnly, ValueError) as err:
            # If EtcdNotDir is raised, then the specified path is a file and
            # thus this is an error.
            log.error("etcd: %s", err)
            return None
        except MaxRetryError as err:
            log.error("etcd: Could not connect to etcd server: %s", err)
            return None
        except Exception as err:  # pylint: disable=broad-except
            log.error("etcd: uncaught exception %s", err)
            raise

        return getattr(result, "dir")

    def ls(self, path):
        ret = {}
        try:
            items = self.read(path)
        except (etcd.EtcdKeyNotFound, ValueError):
            return {}
        except etcd.EtcdConnectionFailed:
            log.error(
                "etcd: failed to perform 'ls' operation on path %s due to connection"
                " error",
                path,
            )
            return None

        # This will find the top level keys only since it's not recursive
        for item in items.children:
            if item.dir is True:
                if item.key == path:
                    continue
                dir_name = f"{item.key}/"
                ret[dir_name] = {}
            else:
                ret[item.key] = item.value
        return {path: ret}

    def delete(self, key, recurse=False, **kwargs):
        recursive = kwargs.pop("recursive", None)
        if recursive is not None:
            salt.utils.versions.warn_until(
                3008,
                "The recursive kwarg has been deprecated, and will be removed "
                "in the Argon release. Please use recurse instead.",
            )
            recurse = recursive
        if kwargs:
            log.warning("Invalid kwargs passed in will not be used: %s", kwargs)

        try:
            if self.client.delete(key, recursive=recurse):
                return True
            else:
                return False
        except (
            etcd.EtcdNotFile,
            etcd.EtcdRootReadOnly,
            etcd.EtcdDirNotEmpty,
            etcd.EtcdKeyNotFound,
            etcd.EtcdConnectionFailed,
            ValueError,
        ) as err:
            log.error("etcd: %s", err)
            return None
        except MaxRetryError as err:
            log.error("etcd: Could not connect to etcd server: %s", err)
            return None
        except Exception as err:  # pylint: disable=broad-except
            log.error("etcd: uncaught exception %s", err)
            raise

    def tree(self, path):
        ret = {}
        try:
            items = self.read(path)
        except (etcd.EtcdKeyNotFound, ValueError):
            return None
        except etcd.EtcdConnectionFailed as err:
            log.error("etcd: %s", err)
            return None

        for item in items.children:
            comps = str(item.key).split("/")
            if item.dir is True:
                if item.key == path:
                    continue
                ret[comps[-1]] = self.tree(item.key)
            else:
                ret[comps[-1]] = item.value
        return ret


class EtcdClientV3(EtcdBase):
    """
    .. versionadded:: 3005

    Since etcd3 has no concept of directories, this class leaves write_directory unimplemented.
    """

    def __init__(
        self,
        opts,
        encode_keys=None,
        encode_values=None,
        raw_keys=False,
        raw_values=False,
        unicode_errors=None,
        **kwargs,
    ):
        if not HAS_ETCD_V3:
            raise EtcdLibraryNotInstalled("Don't have etcd3-py, need to install it.")
        log.debug("etcd_util has the libraries needed for etcd v3")

        super().__init__(opts, **kwargs)

        if self.conf.get("etcd.require_v2", True):
            raise IncompatibleEtcdRequirements("Can't create v3 with a v2 requirement")

        self.encode_keys = encode_keys or self.conf.get("etcd.encode_keys", False)
        self.encode_values = encode_values or self.conf.get("etcd.encode_values", True)
        self.raw_keys = raw_keys or self.conf.get("etcd.raw_keys", False)
        self.raw_values = raw_values or self.conf.get("etcd.raw_values", False)
        self.unicode_errors = unicode_errors or self.conf.get(
            "etcd.unicode_errors", "surrogateescape"
        )

        # etcd3-py uses verify instead of ca_cert
        self.xargs["verify"] = self.xargs.pop("ca_cert", None)
        self.client = etcd3.Client(host=self.host, port=self.port, **self.xargs)

    def _maybe_decode_key(self, key, **extra_kwargs):
        extra_kwargs.setdefault("unicode_errors", self.unicode_errors)
        if self.encode_keys:
            key = salt.utils.msgpack.loads(key, raw=self.raw_keys, **extra_kwargs)
        elif not self.raw_keys and isinstance(key, bytes):
            key = key.decode(encoding="UTF-8", errors=self.unicode_errors)
        return key

    def _maybe_encode_key(self, key, **extra_kwargs):
        extra_kwargs.setdefault("unicode_errors", self.unicode_errors)
        if self.encode_keys:
            key = salt.utils.msgpack.dumps(key, **extra_kwargs)
        return key

    def _maybe_decode_value(self, value, **extra_kwargs):
        extra_kwargs.setdefault("unicode_errors", self.unicode_errors)
        if self.encode_values:
            value = salt.utils.msgpack.loads(value, raw=self.raw_values, **extra_kwargs)
        elif not self.raw_values and isinstance(value, bytes):
            value = value.decode(encoding="UTF-8", errors=self.unicode_errors)
        return value

    def _maybe_encode_value(self, value, **extra_kwargs):
        extra_kwargs.setdefault("unicode_errors", self.unicode_errors)
        if self.encode_values:
            value = salt.utils.msgpack.dumps(value, **extra_kwargs)
        return value

    def _decode_kv(self, kv):
        try:
            kv.key = self._maybe_decode_key(kv.key)
            kv.value = self._maybe_decode_value(kv.value)
        except AttributeError as err:
            log.error("etcd3 decoding error: %s", err)
        return kv

    def watch(self, key, recurse=False, timeout=0, start_revision=None, **kwargs):
        index = kwargs.pop("index", None)
        if index is not None:
            salt.utils.versions.warn_until(
                3008,
                "The index kwarg has been deprecated, and will be removed "
                "in the Argon release. Please use start_revision instead.",
            )
            start_revision = index
        if kwargs:
            log.warning("Invalid kwargs passed in will not be used: %s", kwargs)

        ret = {"key": key, "value": None, "changed": False, "mIndex": 0, "dir": False}
        result = self.read(
            key,
            recurse=recurse,
            wait=True,
            timeout=timeout,
            start_revision=start_revision,
        )
        if result is not None:
            ret["key"] = result.key
            ret["value"] = result.value
            ret["mIndex"] = getattr(result, "mod_revision", 0)
            ret["changed"] = True
        else:
            return None
        return ret

    def get(self, key, recurse=False):
        if not recurse:
            result = self.read(key)
            if isinstance(result, list):
                return result.pop().value
            return None
        return self.tree(key)

    def read(
        self,
        key,
        recurse=False,
        wait=False,
        timeout=None,
        start_revision=None,
        **kwargs,
    ):
        recursive = kwargs.pop("recursive", None)
        wait_index = kwargs.pop("waitIndex", None)
        if recursive is not None:
            salt.utils.versions.warn_until(
                3008,
                "The recursive kwarg has been deprecated, and will be removed "
                "in the Argon release. Please use recurse instead.",
            )
            recurse = recursive
        if wait_index is not None:
            salt.utils.versions.warn_until(
                3008,
                "The waitIndex kwarg has been deprecated, and will be removed "
                "in the Argon release. Please use start_revision instead.",
            )
            start_revision = wait_index
        if kwargs:
            log.warning("Invalid kwargs passed in will not be used: %s", kwargs)

        if not wait:
            try:
                result = self.client.range(self._maybe_encode_key(key), prefix=recurse)
                kvs = getattr(result, "kvs", None)
                if kvs is None:
                    log.error("etcd3 read: No values found for key %s", key)
                else:
                    for kv in kvs:
                        kv = self._decode_kv(kv)
                return kvs
            except Exception as err:  # pylint: disable=W0703
                log.error("etcd3 read: %s", err)
                return None
        else:
            try:
                watcher = self.client.Watcher(
                    key=self._maybe_encode_key(key),
                    prefix=recurse,
                    start_revision=start_revision,
                )
                watch_event = watcher.watch_once(timeout=timeout)
                return self._decode_kv(watch_event)
            except Exception as err:  # pylint: disable=W0703
                log.error("etcd3 watch: %s", err)
                return None

    def update(self, fields, path=""):
        if not isinstance(fields, dict):
            log.error("etcd.update: fields is not type dict")
            return None
        fields = self._flatten(fields, path)
        keys = {}
        for k, v in fields.items():
            if isinstance(v, dict):
                # Not hard failing here so we don't get a partial update
                log.warning("etcd3 has no concept of directories, skipping key %s", k)
                continue
            keys[k] = self.write(k, v)
        return keys

    def write_file(self, key, value, ttl=None):
        if ttl:
            lease = self.client.Lease(ttl=ttl)
            lease.grant()  # We need to explicitly grant the lease
            self.client.put(
                self._maybe_encode_key(key),
                self._maybe_encode_value(value),
                lease=lease.ID,
            )
        else:
            self.client.put(
                self._maybe_encode_key(key), self._maybe_encode_value(value)
            )
        return self.get(key)

    def write_directory(self, key, value, ttl=None):
        raise Etcd3DirectoryException("etcd3 does not have directories")

    def ls(self, path):
        ret = {}
        tree = self.tree(path)

        # Here we will simulate directories because v3 does not have them
        if tree is None:
            return {}
        else:
            sep = "/" if not self.raw_keys else b"/"
            path = (
                path
                if not self.raw_keys or isinstance(path, bytes)
                else path.encode("UTF-8", errors=self.unicode_errors)
            )

            for key, value in tree.items():
                if not path.endswith(sep):
                    ret_key = path + sep + key
                else:
                    ret_key = path + key
                if isinstance(value, dict):
                    ret[ret_key + sep] = {}
                else:
                    ret[ret_key] = value

        return {path: ret}

    def delete(self, key, recurse=False, **kwargs):
        recursive = kwargs.pop("recursive", None)
        if recursive is not None:
            salt.utils.versions.warn_until(
                3008,
                "The recursive kwarg has been deprecated, and will be removed "
                "in the Argon release. Please use recurse instead.",
            )
            recurse = recursive

        if kwargs:
            log.warning("Invalid kwargs passed in will not be used: %s", kwargs)
        result = self.client.delete_range(self._maybe_encode_key(key), prefix=recurse)
        if hasattr(result, "deleted"):
            return True if result.deleted else None
        return False

    def _expand_recurse(self, key, value, dest):
        """
        Helper for _expand
        """
        sep = "/" if not self.raw_keys else b"/"
        outer, *inner = key.lstrip(sep).split(sep, 1)
        if inner:
            if outer not in dest:
                dest[outer] = {}
            self._expand_recurse(inner[0], value, dest[outer])
        else:
            dest[outer] = value

    def _expand(self, kvs):
        """
        This does the opposite of EtcdBase._flatten

        For example, it will convert...

        {
            "/key1": "value1",
            "/key2/subkey1": "subvalue1",
            "/key2/subkey2": "subvalue2",
        }

        to...

        {
            "key1": "value1",
            "key2": {
                "subkey1": "subvalue1",
                "subkey2": "subvalue2",
            }
        }
        """
        dest = {}
        for key, value in kvs.items():
            self._expand_recurse(key, value, dest)
        return dest

    def tree(self, path):
        items = self.read(path, recurse=True)
        if items is None:
            return None
        sep = "/" if not self.raw_keys else b"/"
        if len(items) == 1 and items[0].key == path:
            kv = items.pop()
            return {kv.key.split(sep)[-1]: kv.value}
        kvs = {kv.key[len(path) :]: kv.value for kv in items}
        return self._expand(kvs)


def _get_etcd_opts(opts, profile=None):
    opts_pillar = opts.get("pillar", {})
    opts_master = opts_pillar.get("master", {})

    opts_merged = {}
    opts_merged.update(opts_master)
    opts_merged.update(opts_pillar)
    opts_merged.update(opts)

    if profile:
        return opts_merged.pop(profile, {})
    else:
        return opts_merged


def get_conn(opts, profile=None, **kwargs):
    """
    Client creation at the module level.

    This is the way users are meant to instantiate a client
    """

    conf = _get_etcd_opts(opts, profile=profile)

    # Figure out which API version they are using...
    use_v2 = conf.get("etcd.require_v2", True)
    if use_v2:
        salt.utils.versions.warn_until(
            3009,
            "etcd API v2 has been deprecated.  It will be removed in "
            "the Potassium release, and etcd API v3 will be the default.",
        )
        client = EtcdClient(conf, has_etcd_opts=True, **kwargs)
        log.debug("etcd_util will be attempting to use etcd API v2: python-etcd")
    else:
        client = EtcdClientV3(conf, has_etcd_opts=True, **kwargs)
        log.debug("etcd_util will be attempting to use etcd API v3: etcd3-py")

    return client


def tree(client, path):
    """
    Module level find tree at the given path.
    """
    return client.tree(path)



import logging

log = logging.getLogger(__name__)


def test_issue_62590(salt_master, salt_minion, salt_cli):

    statepy = """
    # _states/test2.py
    import logging
    log = logging.getLogger(__name__)

    def call_another(name, m_name, **kwargs):
        ret = __states__[m_name](name, **kwargs)
        log.info(f'{__opts__["test"]}: {ret}')
        return ret
    """
    statesls = """
    run indirect:
      test2.call_another:
        - m_name: test.succeed_with_changes

    run prereq:
      test2.call_another:
        - m_name: test.succeed_with_changes

    nop:
      test.nop:
        - prereq:
          - run prereq
    """
    with salt_master.state_tree.base.temp_file(
        "_states/test2.py", statepy
    ), salt_master.state_tree.base.temp_file("test_62590.sls", statesls):
        ret = salt_cli.run("saltutil.sync_all", minion_tgt=salt_minion.id)
        assert ret.returncode == 0
        ret = salt_cli.run("state.apply", "test_62590", minion_tgt=salt_minion.id)
        assert ret.returncode == 0
        assert "Success!" == ret.data["test_|-nop_|-nop_|-nop"]["comment"]


def test_failing_sls(salt_master, salt_minion, salt_cli, caplog):
    """
    Test when running state.sls and the state fails.
    When the master stores the job and attempts to send
    an event a KeyError was previously being logged.
    This test ensures we do not log an error when
    attempting to send an event about a failing state.
    """
    statesls = """
    test_state:
      test.fail_without_changes:
        - name: "bla"
    """
    with salt_master.state_tree.base.temp_file("test_failure.sls", statesls):
        ret = salt_cli.run("state.sls", "test_failure", minion_tgt=salt_minion.id)
        for message in caplog.messages:
            assert "Event iteration failed with" not in message


def test_failing_sls_compound(salt_master, salt_minion, salt_cli, caplog):
    """
    Test when running state.sls in a compound command and the state fails.
    When the master stores the job and attempts to send
    an event a KeyError was previously being logged.
    This test ensures we do not log an error when
    attempting to send an event about a failing state.
    """
    statesls = """
    test_state:
      test.fail_without_changes:
        - name: "bla"
    """
    with salt_master.state_tree.base.temp_file("test_failure.sls", statesls):
        ret = salt_cli.run(
            "state.sls,cmd.run", "test_failure,ls", minion_tgt=salt_minion.id
        )
        for message in caplog.messages:
            assert "Event iteration failed with" not in message



"""
integration tests for mac_ports
"""

import pytest

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
    pytest.mark.skip_if_binaries_missing("port"),
]


@pytest.fixture(scope="module")
def pkg(modules):
    return modules.pkg


@pytest.fixture
def uninstalled_pkg_name(pkg):
    pkgname = installed_pkg_name
    try:
        pkg.refresh_db()
        yield pkgname
    finally:
        if pkgname in pkg.list_pkgs():
            pkg.remove(pkgname)


@pytest.fixture
def installed_pkg_name(uninstalled_pkg_name):
    pkg.install(uninstalled_pkg_name)
    return uninstalled_pkg_name


@pytest.fixture(scope="function", autouse=True)
def _setup_teardown_vars(pkg):
    AGREE_INSTALLED = False
    try:
        ret = pkg.list_pkgs()
        AGREE_INSTALLED = installed_pkg_name in ret
        pkg.refresh_db()
        yield
    finally:
        if AGREE_INSTALLED:
            pkg.remove(installed_pkg_name)


def test_list_pkgs(pkg, installed_pkg_name):
    """
    Test pkg.list_pkgs
    """
    pkg_list_ret = pkg.list_pkgs()
    assert isinstance(pkg_list_ret, dict)
    assert installed_pkg_name in pkg_list_ret


def test_latest_version(pkg, installed_pkg_name):
    """
    Test pkg.latest_version
    """
    result = pkg.latest_version(installed_pkg_name, refresh=False)
    assert isinstance(result, dict)
    assert installed_pkg_name in result.data


def test_remove(pkg, installed_pkg_name):
    """
    Test pkg.remove
    """
    ret = pkg.remove(installed_pkg_name)
    assert isinstance(ret, dict)
    assert installed_pkg_name in ret


@pytest.mark.destructive_test
def test_install(pkg, uninstalled_pkg_name):
    """
    Test pkg.install
    """
    ret = pkg.install(uninstalled_pkg_name)
    assert isinstance(ret, dict)
    assert uninstalled_pkg_name in ret


def test_list_upgrades_type(pkg):
    """
    Test pkg.list_upgrades return type
    """
    ret = pkg.list_upgrades(refresh=False)
    assert isinstance(ret, dict)


def test_upgrade_available(pkg, installed_pkg_name):
    """
    Test pkg.upgrade_available
    """
    ret = pkg.upgrade_available(installed_pkg_name, refresh=False)
    assert not ret.data


def test_refresh_db(pkg):
    """
    Test pkg.refresh_db
    """
    ret = pkg.refresh_db()
    assert ret


def test_upgrade(pkg):
    """
    Test pkg.upgrade
    """
    ret = pkg.upgrade(refresh=False)
    assert isinstance(ret, dict)
    assert ret.data["result"]
